try:
    import spacy
    import numpy as np
    import pandas as pd
    from spacy.util import minibatch, compounding
    from fastprogress.fastprogress import master_bar, progress_bar
    from .._utils.common import _get_emd_path
    from .._utils.text_data import copy_metrics
    from ..models._codetemplate import entity_recognizer_placeholder

    HAS_SPACY = True
except:
    HAS_SPACY = False

import json, logging
from pathlib import Path
import random, os
import datetime
from copy import deepcopy
from collections.abc import Iterable
import tempfile
from .._utils._ner_utils import even_mults, _timelapsed
from ..models._arcgis_model import ArcGISModel, _create_zip


class _SpacyEntityRecognizer(ArcGISModel):
    """
    Creates an entity recognition model to extract text entities from unstructured text documents.
    Based on Spacy's `EntityRecognizer <https://spacy.io/api/entityrecognizer>`_

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Requires data object returned from
                            :meth:`~arcgis.learn.prepare_data`  function.
    ---------------------   -------------------------------------------
    lang                    Optional string. Language-specific code,
                            named according to the language’s `ISO code <https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes>`_
                            The default value is 'en' for English.
    =====================   ===========================================

    :return: ``_SpacyEntityRecognizer`` Object
    """

    def __init__(self, data=None, lang="en", *args, **kwargs):
        super().__init__(data)
        self._code = entity_recognizer_placeholder
        self._emd_template = {}
        self.model_dir = None
        self.saved_model_dir = None
        self.model = spacy.blank(lang)
        self.ner = self.model.create_pipe("ner")
        self.model.add_pipe(self.ner, last=True)
        self._address_tag = "Address"  # Defines the default addres field
        self.entities = (
            None  # Stores all the entity names from the training data into a list
        )
        self._has_address = (
            False  # Flag to identify if the training data has any address
        )
        self._trained = False  # Flag to check if model has been trained
        self.lang = lang
        self.optimizer = self.model.begin_training()
        if data:
            self._is_empty = False
            self._address_tag = data._address_tag
            self._has_address = data._has_address
            self.path = data.working_dir
            self.data = data
            self.train_ds = data.train_ds
            self.val_ds = data.val_ds
            for ent in data.entities:
                if ent not in self.ner.labels:
                    self.model.entity.add_label(ent)
        else:
            self._is_empty = True
            self.train_ds = None
            self.val_ds = None
            self.path = "."
            self._is_empty = True
        self.learn = self
        self.recorder = Recorder()
        self.model_characteristics_folder = "ModelCharacteristics"

        pretrained_path = kwargs.get("pretrained_path", None)
        if pretrained_path is not None:
            pretrained_path = str(_get_emd_path(pretrained_path))
            self.load(pretrained_path)

    def lr_find(self, allow_plot=True):
        """
        Runs the Learning Rate Finder, and displays the graph of it's output.
        Helps in choosing the optimum learning rate for training the model.
        """
        self._check_requisites()

        start_lr = 1e-6
        end_lr = 10
        num_it = 10
        smoothening = 4

        with tempfile.TemporaryDirectory(prefix="arcgisTemp_") as _tempfolder:
            checkpoint_path = os.path.join(_tempfolder, "tmp")
            if not os.path.exists(os.path.dirname(checkpoint_path)):
                os.mkdir(os.path.dirname(checkpoint_path))
            self.model.to_disk(checkpoint_path)  # caches the current model state
            if self._trained:
                temp_optimizer = (
                    self.optimizer
                )  # preserving the current state of the model for later load
            trained = (
                self._trained
            )  # preserving the current state of the model for later load
            recorder = deepcopy(
                self.recorder
            )  # preserving the current state of the model for later load
            self.recorder.losses, self.recorder.val_loss, self.recorder.lrs = (
                [],
                [],
                [],
            )  # resetting the recorder
            lrs = even_mults(start_lr, end_lr, 14)
            epochs = int(
                np.ceil(num_it / (len(self.data.train_ds) / self.data.batch_size))
            )
            self.fit(
                lr=list(lrs), epochs=epochs * len(lrs), from_lr_find=True, num_it=num_it
            )
            from IPython.display import clear_output

            clear_output()

            N = smoothening  # smoothening factor
            self.recorder.losses = np.convolve(
                self.recorder.losses, np.ones((N,)) / N, mode="valid"
            ).tolist()
            self.recorder.lrs = np.convolve(
                self.recorder.lrs, np.ones((N,)) / N, mode="valid"
            ).tolist()
            lr, index = self._find_lr(
                losses_skipped=0, trailing_losses_skipped=0, section_factor=2
            )

            if allow_plot:
                self._show_lr_plot(index, losses_skipped=0, trailing_losses_skipped=1)
            self._trained = trained
            self.recorder = recorder
            import spacy

            self.model = spacy.load(checkpoint_path)
        return lr

    def unfreeze(self):
        """
        Not implemented for this model.
        """
        logging.error(
            "unfreeze() is not implemented for EntityRecognizer model with spaCy backbone."
        )

    def freeze(self):
        """
        Not implemented for this model.
        """
        logging.error(
            "freeze() is not implemented for EntityRecognizer model with spaCy backbone."
        )

    def fit(
        self,
        epochs=20,
        lr=None,
        one_cycle=True,
        early_stopping=False,
        checkpoint=True,
        **kwargs,
    ):
        """
        Trains an EntityRecognition model for 'n' number of epochs..

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        epoch                   Optional integer. Number of times the model will train
                                on the complete dataset.
        ---------------------   -------------------------------------------
        lr                      Optional float. Learning rate
                                to be used for training the model.
        ---------------------   -------------------------------------------
        one_cycle               Not implemented for this model.
        ---------------------   -------------------------------------------
        early_stopping          Not implemented for this model.
        ---------------------   -------------------------------------------
        checkpoint              Not implemented for this model.
        =====================   ===========================================
        """
        self._check_requisites()
        if (
            lr is None
        ):  # searching for the optimal learning rate when no learning rate is provided
            print("Finding optimum learning rate")
            lr = self.lr_find(allow_plot=False)

        if kwargs.get("from_lr_find", False) is False and isinstance(lr, slice):
            lr = lr.stop
            error_message = (
                "Passing slice of floats as `lr` value is not supported for models with `spacy` backbone."
                f" Picking up the highest value - `{lr}` of the slice as the learning rate."
            )
            logging.warning(error_message)

        if self.train_ds is None:
            return logging.warning("Cannot fit the model on empty data.")

        TRAIN_DATA = self.train_ds.data
        VAL_DATA = self.val_ds.data
        nlp = self.model

        if (
            "ner" not in nlp.pipe_names
        ):  # create the built-in pipeline components and add them to the pipeline
            # spacy.require_gpu()
            self.ner = nlp.create_pipe(
                "ner"
            )  # nlp.create_pipe works for built-ins that are registered with spaCy
            nlp.add_pipe(self.ner, last=True)

        for _, annotations in TRAIN_DATA:  # adding labels
            for ent in annotations.get("entities"):
                if ent[2] not in self.ner.labels:
                    self.ner.add_label(ent[2])

        other_pipes = [
            pipe for pipe in nlp.pipe_names if pipe != "ner"
        ]  # get names of other pipes to disable them during training
        with nlp.disable_pipes(*other_pipes):  # only train NER
            if not nlp.vocab.vectors.name:
                nlp.vocab.vectors.name = "spacy_pretrained_vectors"

            batch_size = self.data.batch_size
            n_iter = len(TRAIN_DATA) // batch_size
            if (
                "from_lr_find" in kwargs
            ):  # 'from_lr_find' kwarg specifies that the fit is call from lr_find.
                epochs_per_lr = epochs / len(lr)
                n_iter = min(kwargs.get("num_it"), n_iter)
                lr_find = True
            else:
                self.optimizer.alpha = lr
                lr_find = False
            mb = master_bar(range(epochs))
            mb.write(
                [
                    "epoch",
                    "losses",
                    "val_loss",
                    "precision_score",
                    "recall_score",
                    "f1_score",
                    "time",
                ],
                table=True,
            )
            losses_list = []

            for itn in mb:
                t_start = datetime.datetime.now()
                if lr_find:
                    self.optimizer.alpha = lr.pop(0)
                    losses_list = []
                    update_recorder = True
                random.shuffle(TRAIN_DATA)
                batches = minibatch(TRAIN_DATA, size=batch_size)
                losses = {}
                epoch_loss = []
                for batch_index in progress_bar(range(n_iter), parent=mb):
                    batch_index += 1
                    batch = next(batches)
                    texts, annotations = zip(*batch)
                    nlp.update(
                        texts, annotations, sgd=self.optimizer, drop=0.35, losses=losses
                    )
                    processed_len = len(batch) * batch_index
                    train_loss = (
                        losses["ner"] / processed_len
                    )  # normalized with processed_len
                    if lr_find:
                        losses_list.append(train_loss)
                    else:  # recording training loss per iteration.
                        epoch_loss.append(train_loss)

                if not lr_find:
                    self.recorder.losses.append(
                        sum(epoch_loss) / n_iter
                    )  # averaging loss per epoch

                if VAL_DATA:
                    if lr_find:
                        VAL_DATA = VAL_DATA[
                            :batch_size
                        ]  # running on a subset of val data incase of lr_find
                    val_batches = minibatch(VAL_DATA, size=batch_size)
                    val_losses = {}
                    val_loss_list = []
                    epoch_loss = []
                    for batch_index, val_batch in enumerate(val_batches):
                        batch_index += 1
                        processed_len_val = batch_size * (batch_index)
                        val_text, val_annotations = zip(*val_batch)
                        nlp.update(
                            val_text, val_annotations, sgd=None, losses=val_losses
                        )
                        val_loss = (
                            val_losses["ner"] / processed_len_val
                        )  # normalized with processed_len_val
                        if lr_find:
                            val_loss_list.append(val_loss)
                        else:  # recording validation loss per iteration.
                            epoch_loss.append(val_loss)
                    if not lr_find:
                        self.recorder.val_loss.append(
                            sum(epoch_loss) / batch_index
                        )  # averaging loss per epoch

                if lr_find:
                    self.recorder.losses.append(np.mean(losses_list))
                    self.recorder.lrs.append(self.optimizer.alpha)
                    self.recorder.val_loss.append(np.min(val_loss_list))
                    update_recorder = False
                    # break the epoch if loss overshoots or all the lrs are tested
                    if (
                        np.mean(losses_list) > 2 * np.min(self.recorder.losses)
                        or len(lr) == 0
                    ):
                        return
                    score = nlp.evaluate(self.val_ds[:batch_size])
                else:
                    score = nlp.evaluate(self.val_ds)
                precision_score, recall_score, f1_score, metrics_per_label = (
                    score.ents_p,
                    score.ents_r,
                    score.ents_f,
                    score.ents_per_type,
                )
                self.recorder.metrics["precision_score"].append(precision_score)
                self.recorder.metrics["recall_score"].append(recall_score)
                self.recorder.metrics["f1_score"].append(f1_score)
                self.recorder.metrics["metrics_per_label"].append(metrics_per_label)
                line = [
                    itn,
                    round(train_loss, 2),
                    round(val_loss, 2),
                    round(precision_score / 100, 2),
                    round(recall_score / 100, 2),
                    round(f1_score / 100, 2),
                    _timelapsed(t_start),
                ]
                line = [str(val) for val in line]
                mb.write(line, table=True)

        if not lr_find:
            self._trained = True
            self.model = nlp
            self.entities = list(self.model.entity.labels)
            self.lr = lr

    def _create_emd(self, path, compute_metrics=True):
        path = Path(path)
        self._emd_template = {}
        self._emd_template["ModelConfiguration"] = "_ner"
        self._emd_template["InferenceFunction"] = "EntityRecognizer.py"
        self._emd_template["ModelFile"] = str(Path(path).name)
        self._emd_template["ModelName"] = type(self).__name__
        self._emd_template["Labels"] = self.model.entity.labels
        self._emd_template["Lang"] = self.lang
        self._emd_template["saved_path"] = str(Path(path))
        if hasattr(self, "lr"):
            self._emd_template["LearningRate"] = str(
                self.lr
            )  # checking if model has lr
        if compute_metrics and len(
            self.recorder.metrics["precision_score"]
        ):  # checking if recorder has metrics
            self._emd_template["metrics"] = json.dumps(
                {
                    "precision_score": [self.recorder.metrics["precision_score"][-1]],
                    "recall_score": [self.recorder.metrics["recall_score"][-1]],
                    "f1_score": [self.recorder.metrics["f1_score"][-1]],
                    "metrics_per_label": [
                        self.recorder.metrics["metrics_per_label"][-1]
                    ],
                }
            )
        if self._has_address:
            self._emd_template["address_tag"] = self._address_tag

        json.dump(
            self._emd_template,
            open(path / Path(path.stem).with_suffix(".emd"), "w"),
            indent=4,
        )
        pathstr = path / Path(path.stem).with_suffix(".emd")
        return path.resolve()

    def save(self, name_or_path, **kwargs):
        """
        Saves the model weights, creates an Esri Model Definition.
        Train the model for the specified number of epochs and using the
        specified learning rates.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Name of the model to save. It
                                stores it at the pre-defined location. If path
                                is passed then it stores at the specified path
                                with model name as directory name. and creates
                                all the intermediate directories.
        =====================   ===========================================
        """
        return self._save(name_or_path, **kwargs)

    def _save_model_characteristics(self, model_characteristics_dir):
        if not os.path.exists(model_characteristics_dir):
            os.makedirs(model_characteristics_dir)

        fig = self.plot_losses(show=False)
        if fig:
            fig.savefig(os.path.join(model_characteristics_dir, "loss_graph.png"))
        from IPython.utils import io

        with io.capture_output() as captured:
            self.show_results().to_html(
                os.path.join(model_characteristics_dir, "results.html")
            )
            metrics = self.metrics_per_label()
            if metrics is not None:  # expecting metrics = None for older saved models.
                metrics.to_html(
                    os.path.join(model_characteristics_dir, "metrics.html")
                )  # writing metrics to html

    def _save(
        self,
        name_or_path,
        zip_files=True,
        save_html=True,
        publish=False,
        gis=None,
        compute_metrics=True,
        **kwargs,
    ):
        temp = self.path
        if not self._trained:
            return logging.error("Model needs to be fitted, before saving.")

        if kwargs.get("save_optimizer", False):
            logging.warning(
                "Setting `save_optimizer` = True will not have any effect on models with `spaCy` backbone"
            )

        if "\\" in name_or_path or "/" in name_or_path:
            path = Path(name_or_path)
            parent_path = path.parent
            name = path.parts[-1]
            self.model_dir = parent_path / name
            if not os.path.exists(self.model_dir):
                os.makedirs(self.model_dir)
        else:
            self.model_dir = Path(self.path) / "models" / name_or_path
            name = name_or_path
            if not os.path.exists(self.model_dir):
                os.makedirs(self.model_dir)

        self.model.to_disk(self.model_dir)
        emd_path = self._create_emd(self.model_dir, compute_metrics=compute_metrics)
        with open(self.model_dir / self._emd_template["InferenceFunction"], "w") as f:
            f.write(self._code)

        if save_html:
            if self._is_empty:
                copy_metrics(
                    self.saved_model_dir,
                    self.model_dir,
                    self.model_characteristics_folder,
                )
            else:
                self._save_model_characteristics(
                    self.model_dir.absolute() / self.model_characteristics_folder
                )
                self._create_html(Path(self.model_dir.absolute() / self.model_dir.stem))

        if zip_files:
            _create_zip(name, str(self.model_dir))

        if publish:
            self._publish_dlpk(
                (emd_path / emd_path.stem).with_suffix(".dlpk"),
                gis=gis,
                overwrite=kwargs.get("overwrite", False),
            )
        print(f"Model has been saved to {str(emd_path)}")
        return emd_path

    def load(self, name_or_path):
        """
        Loads a saved EntityRecognition model from disk.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Path of the emd file.
        =====================   ===========================================
        """
        if "\\" in str(name_or_path) or "/" in str(name_or_path):
            name_or_path = str(_get_emd_path(str(name_or_path)))
            name_or_path = name_or_path
            model_path = Path(name_or_path).parent
        else:
            model_path = Path(self.path) / "models" / name_or_path
            name_or_path = (
                Path(self.path) / "models" / name_or_path / f"{name_or_path}.emd"
            )

        with open(name_or_path, "r", encoding="utf-8") as f:
            emd = f.read()
        emd = json.loads(emd)
        address_tag = emd.get("address_tag")
        if address_tag:
            self._has_address = True
            self._address_tag = address_tag
        self.model = spacy.load(model_path)
        self.ner = self.model.get_pipe("ner")
        self._trained = True
        self.entities = list(self.model.entity.labels)
        self.model_dir = Path(name_or_path).parent.resolve()
        self.recorder = Recorder()
        if emd.get("metrics"):
            self.recorder.metrics = json.loads(emd.get("metrics"))
        self.saved_model_dir = deepcopy(self.model_dir)

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates an :class:`~arcgis.learn.text.EntityRecognizer` from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Esri Model Definition
                                file.
        ---------------------   -------------------------------------------
        data                    Required DatabunchNER object or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data` function or None for
                                inferencing.

        =====================   ===========================================

        :return: :class:`~arcgis.learn.text.EntityRecognizer` Object
        """
        emd_path = Path(emd_path)
        ner = cls(data=data)
        ner.load(emd_path)
        ner._trained = True
        ner.entities = list(ner.model.entity.labels)
        return ner

    def _post_process_non_address_df(self, unprocessed_df):
        """
        This function post processes the output dataframe from extract_entities function and returns a processed dataframe.
        """
        processed_df = pd.DataFrame(columns=unprocessed_df.columns)
        for col in unprocessed_df.columns:  # converting all list columns to string
            if (
                pd.Series(filter(lambda x: x != "", unprocessed_df[col]))
                .apply(isinstance, args=([str]))
                .sum()
                == 0
            ):  # split if this condition
                processed_df[col] = unprocessed_df[col].apply(
                    ",".join
                )  # join the list to string and copy to the processed df
            else:
                processed_df[col] = unprocessed_df[col]  # copy to the processed df
        return processed_df

    def _post_process_address_df(self, unprocessed_df, drop):
        """
        This function post processes the output dataframe from extract_entities function and returns a processed dataframe with cleaned up missed detections.
        """
        address_tag = self._address_tag
        processed_df = pd.DataFrame(
            columns=unprocessed_df.columns
        )  # creating an empty processed dataframe
        for i, adds in unprocessed_df[
            address_tag
        ].items():  # duplicating rows with multiple addresses to be one row per address
            if len(adds) > 0:  # adding data for address documents
                for j, add in enumerate(adds):
                    curr_index = len(processed_df)
                    processed_df.loc[curr_index] = unprocessed_df.loc[i]
                    processed_df.loc[curr_index, address_tag] = add
            else:  # adding data for non-address documents
                curr_index = len(processed_df)
                processed_df.loc[curr_index] = unprocessed_df.loc[i]
                processed_df.loc[curr_index, address_tag] = ""
        drop_ids = []

        for i, add in processed_df[address_tag].items():
            if len(add.split(" ")) < 2:
                drop_ids.append(i)
        del unprocessed_df

        if drop:  # flag for dropping/not-dropping documents without address.
            processed_df.drop(drop_ids, inplace=True)
        cols = processed_df.columns
        processed_df.reset_index(drop=True, inplace=True)

        for col in processed_df.columns:  # converting all list columns to string
            if (
                col != address_tag
                and pd.Series(filter(lambda x: x != "", processed_df[col]))
                .apply(isinstance, args=([str]))
                .sum()
                == 0
            ):  ## split if this condition
                processed_df[col] = processed_df[col].apply(
                    ",".join
                )  # join the list to strind and copy to the processed df
            else:
                processed_df[col] = processed_df[col]  # copy to the processed df
        return processed_df

    def _extract_entities_text(self, text):
        """
        This function extracts entities from a string"

        Arguments:
        text:str

        Returns:
        spacy's doc object

        Example of how to visualize the results:
        [(ent.label_, ent.text) for ent in doc_object.ents]
        """
        return self.model(text)

    def extract_entities(self, text_list, drop=True, show_progress=True, **kwargs):
        """
        Extracts the entities from [documents in the mentioned path or text_list].

        Field defined as 'address_tag' in :meth:`~arcgis.learn.prepare_data`  function's class mapping
        attribute will be treated as a location. In cases where trained model extracts
        multiple locations from a single document, that document will be replicated
        for each location in the resulting dataframe.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        text_list               Required string(path) or list(documents).
                                List of documents for entity extraction OR
                                path to the documents.
        ---------------------   -------------------------------------------
        drop                    Optional bool.
                                If documents without address needs to be
                                dropped from the results.
        ---------------------   -------------------------------------------
        batch_size              Optional integer. Number of items to process
                                at once. (Reduce it if getting CUDA Out of Memory
                                Errors). Default is set to 4.
                                Not applicable for models with `spaCy` backbone.
        =====================   ===========================================

        :return: Pandas DataFrame
        """

        if self._trained:
            if isinstance(text_list, list):
                item_list = pd.Series(text_list)

            elif isinstance(text_list, str):
                item_names = os.listdir(text_list)
                item_list = pd.Series()
                text = []
                skipped_docs = []
                for item_name in item_names:
                    try:
                        with open(
                            f"{text_list}/{item_name}",
                            "r",
                            encoding="utf-16",
                            errors="ignore",
                        ) as f:
                            item_list[item_name] = f.read()
                    except:
                        try:
                            with open(
                                f"{text_list}/{item_name}",
                                "r",
                                encoding="utf-8",
                                errors="ignore",
                            ) as f:
                                item_list[item_name] = f.read()
                        except:
                            skipped_docs.append(item_name)
                if len(skipped_docs):
                    print(
                        "Unable to read the following documents ",
                        ", ".join(skipped_docs),
                    )

            # if self._address_tag not in self.entities and self._has_address==True:
            #     return logging.warning('Model\'s address tag does not match with any field in your data, one of the below steps could resolve your issue:\n\
            #         1. Set address tag to the address field in your data [your_model._address_tag=\'your_address_field\']\n\
            #         2. If your data does not have any address field set _has_address=False [your_model._has_address=False]')
            data_list = []
            for i, item in progress_bar(list(item_list.items()), display=show_progress):
                doc = self._extract_entities_text(
                    item
                )  # predicting entities using entity_extractor model
                text = doc.text
                tmp_ents = {}
                for ent in doc.ents:  # Preparing a dataframe from results
                    if tmp_ents.get(ent.label_) == None:
                        tmp_ents[ent.label_] = [] + [ent.text]
                    else:
                        tmp_ents[ent.label_].extend([ent.text])

                tmp_ents["TEXT"] = text
                if isinstance(i, Iterable):  # For test documents
                    tmp_ents["Filename"] = i
                else:  # for show_results()
                    tmp_ents["Filename"] = "Example_" + str(i)

                data_list.append(tmp_ents)

            df = pd.DataFrame(data_list, columns=["TEXT", "Filename"] + self.entities)
            df.fillna("", inplace=True)
            if self._has_address:
                df = self._post_process_address_df(
                    df, drop
                )  # Post processing the dataframe
            else:
                df = self._post_process_non_address_df(
                    df
                )  # Post processing the dataframe
            return df.reset_index(drop="True")
        else:
            return logging.error("Model needs to be fitted, before extraction.")

    def show_results(self, ds_type="valid"):
        """
        Runs entity extraction on a random batch from the mentioned ds_type.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        ds_type                 Optional string, defaults to valid.
        =====================   ===========================================

        :return: Pandas DataFrame
        """

        if not self._trained:
            return logging.warning("This model has not been trained")
        """
        Make predictions on a batch of documents from specified ds_type.
        ds_type:['valid'|'train] 
        """
        # if self._address_tag not in self.entities and self._has_address == True:
        #     return logging.warning('Model\'s address tag does not match with any field in your data, one of the below steps could resolve your issue:\n\
        #         1. Set address tag to the address field in your data [your_model._address_tag=\'your_address_field\']\n\
        #         2. If your data does not have any address field set _has_address=False [your_model._has_address=False]')
        if self._is_empty:
            return logging.warning("This model does not have data.")

        if ds_type.lower() == "valid":
            xs = self.val_ds._random_batch(self.val_ds.x)
            return self.extract_entities(xs, show_progress=False)
        elif ds_type.lower() == "train":
            xs = self.train_ds._random_batch(self.train_ds.x)
            return self.extract_entities(xs, show_progress=False)
        else:
            print("Please provide a valid ds_type:['valid'|'train']")

    def precision_score(self):
        if self._trained:
            precision_pct = self.recorder.metrics["precision_score"][-1]
            precision = round(precision_pct / 100, 2)
            return precision
        else:
            return logging.warning("This model has not been trained")

    def recall_score(self):
        if self._trained:
            recall_pct = self.recorder.metrics["recall_score"][-1]
            recall = round(recall_pct / 100, 2)
            return recall
        else:
            return logging.warning("This model has not been trained")

    def f1_score(self):
        if self._trained:
            f1_pct = self.recorder.metrics["f1_score"][-1]
            f1 = round(f1_pct / 100, 2)
            return f1
        else:
            return logging.warning("This model has not been trained")

    def metrics_per_label(self):
        if self._trained:  # for saving old(before metrics were implemented) models.
            if not len(self.recorder.metrics["metrics_per_label"]):
                return None

            metrics_df = pd.DataFrame(
                self.recorder.metrics["metrics_per_label"][-1]
            ).transpose()
            metrics_df.columns = ["Precision_score", "Recall_score", "F1_score"]
            metrics_df = metrics_df.apply(lambda x: round(x / 100, 2))
            return metrics_df
        else:
            return logging.warning("This model has not been trained")

    def plot_losses(self, show=True):
        """
        Plot training and validation losses.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        show                    Optional bool. Defaults to True
                                If set to False, figure will not be plotted
                                but will be returned, when set to True function
                                will plot the figure and return nothing.
        =====================   ===========================================

        :return: `matplotlib.figure.Figure <https://matplotlib.org/stable/api/figure_api.html#matplotlib.figure.Figure>`_
        """
        self._check_requisites()

        if not len(self.recorder.losses):  # return none if the recorder is empty
            raise Exception("Model needs to be fitted, before saving.")
        import numpy as np
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(1, 1)
        N = max(1, len(self.recorder.losses) // 10)  # smooth with a factor of N
        ax.plot(
            np.convolve(self.recorder.losses, np.ones((N,)) / N, mode="valid"),
            label="Train",
        )
        ax.plot(
            np.convolve(self.recorder.val_loss, np.ones((N,)) / N, mode="valid"),
            label="Validation",
        )
        ax.set_ylabel("Loss")
        ax.set_xlabel("Epochs")
        ax.legend()

        if not show:
            plt.close()
            return fig
        else:
            plt.show()

    def _check_requisites(self):
        if getattr(self, "_is_empty", False):
            raise Exception("Can't call this function without data.")


class Recorder:
    def __init__(self):
        self.lrs = []
        self.losses = []
        self.val_loss = []
        self.metrics = {
            "precision_score": [],
            "recall_score": [],
            "f1_score": [],
            "metrics_per_label": [],
        }



from __future__ import annotations
import csv
import json
import logging
import tempfile
import datetime as _dt
from datetime import datetime
from arcgis.gis.kubernetes._admin._base import _BaseKube
from typing import Dict, Any, Optional, List, Union

_log = logging.getLogger()


########################################################################
class LogManager(_BaseKube):
    """
    Helper class for the management of Kubernetes logs by administrators.

    Logs are the transaction records written by the various components
    of ArcGIS Server.  You can query the logs, change various log settings,
    and check error messages for helping to determine the nature of an issue.

    """

    _url = None
    _con = None
    _json_dict = None
    _json = None

    # ----------------------------------------------------------------------
    def __init__(self, url, gis, initialize=False):
        """Constructor


        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        url                    Required string. The machine URL.
        ------------------     --------------------------------------------------------------------
        gis                    Optional string. The GIS or Server object..
        ==================     ====================================================================

        """
        connection = gis
        super(LogManager, self).__init__(gis=gis, url=url)
        self._url = url
        if hasattr(connection, "_con"):
            self._con = connection._con
        if initialize:
            self._init(connection)

    # ----------------------------------------------------------------------
    def __str__(self):
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    def __repr__(self):
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    def clean(self, start_time=None, end_time=None, level=None):
        """
        Deletes all the log files on all server machines in the site. This is an irreversible
        operation.

        This operation forces the server to clean the logs, which has the effect of freeing up disk space. However, it is not required that you invoke this operation because the server periodically purges old logs.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        start_time          Optional String. The date associated with a log, in timestamp format
                            (yyyy-mm-ddThh:mm:ss). If specified, logs created after this time
                            will be deleted. When the endTime parameter is also defined, only
                            logs created between the date range will be deleted. When `log_level`
                            is added to the request with a date range specified, only logs that
                            match the selected level and were created within the defined date
                            range will be deleted.
        ---------------     --------------------------------------------------------------------
        end_time            Optional String. The date associated with a log, in timestamp
                            format (yyyy-mm-ddThh:mm:ss). If specified, logs created before this
                            time will be deleted. When the startTime parameter is also defined,
                            only logs created between the date range will be deleted. When
                            `log_level` is added to the request with a date range specified,
                            only logs that match the selected level and were created within the
                            defined date range will be deleted.
        ---------------     --------------------------------------------------------------------
        level               Optional String. The level of logs that will be deleted. If no
                            option is selected, all log messages will be deleted.

                            Values: SEVERE | WARNING | INFO | FINE | VERBOSE | DEBUG

        ===============     ====================================================================


        :return:
           A boolean indicating success (True) or failure (False).

        """
        params = {
            "f": "json",
        }
        if start_time:
            params["startTime"] = start_time
        if end_time:
            params["endTime"] = end_time
        if level:
            params["logLevel"] = level
        url = "{}/clean".format(self._url)
        res = self._con.post(path=url, postdata=params)
        if "status" in res:
            return res["status"] == "success"
        elif "success" in res:
            return res["success"] == "true"
        return res

    # ----------------------------------------------------------------------
    def refresh_index(self) -> bool:
        """
        Recreates the log indexes and can be used to troubleshoot issues
        related to accessing logs, such as if new logs are not being
        generated or if existing logs are unavailable.


        :return: Boolean. True if successful else False.

        """
        params = {"f": "json"}
        url = self._url + "/settings/updateLogIndex"
        return self._con.post(url, params).get("status", "failed") == "success"

    # ----------------------------------------------------------------------
    @property
    def settings(self):
        """Gets the current log settings."""
        params = {"f": "json"}
        url = self._url + "/settings"
        try:
            return self._con.get(url, params)
        except:
            return ""

    # ----------------------------------------------------------------------
    def export(
        self,
        query: str | None = None,
        start_time: _dt.datetime | None = None,
        end_time: _dt.datetime | None = None,
        level: str = "WARNING",
        log_code: str | None = None,
        users: list[str] | None = None,
        request_ids: str | None = None,
        service_types: str | None = None,
        source: str | None = None,
        stack_traces: bool = False,
        out_folder: str | None = None,
    ) -> str:
        """
        The export operation exports organization logs based on either query
        or search parameters. Using the query filter parameters, you can
        aggregate and filter through logs for your deployment. Using the
        search parameters, you can search for specific log records. Once
        completed, a .zip file of the exported logs is uploaded to the
        uploads directory, which can be downloaded from the URL provided
        with the success response. If necessary, the export operation can
        be invoked multiple times to acquire additional logs.


        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        query                  Optional string. The search terms used to query your organization's
                               logs. This parameter supports keywords (for example, completed) and
                               phrases (for example, completed successfully).
        ------------------     --------------------------------------------------------------------
        start_time             Optional datetime. The oldest time to query logs against, formatted as
                               either a timestamp (yyyy-mm-ddThh:mm:ss) or milliseconds from epoch.
                               The default is the beginning of all recorded logs.
        ------------------     --------------------------------------------------------------------
        end_time               Optional datetime. The most recent time to query against, formatted as
                               either a timestamp (yyyy-mm-ddThh:mm:ss) or milliseconds from epoch.
                               The default is the current date.
        ------------------     --------------------------------------------------------------------
        level                  Optional string. Gets only the records with a log level at or more
                               severe than the level declared here. Can be one of (in severity
                               order): DEBUG, VERBOSE, FINE, INFO, WARNING, SEVERE. The
                               default is WARNING.
        ------------------     --------------------------------------------------------------------
        log_code               Optional String. Specifies the log codes assigned to the logs. When
                               specified, query will return logs associated with those codes.
        ------------------     --------------------------------------------------------------------
        users                  Optional list[str]. The username(s) of a user within the organization
                               that can be used to further filter log results.

                               example:
                               >>> users = ["user1", "user2"]
                               >>> logs.export(users=users)

        ------------------     --------------------------------------------------------------------
        request_ids            Optional String. An ID assigned to a specific server event.
        ------------------     --------------------------------------------------------------------
        service_types          Optional String. The service type of a service within the
                               organization that can be used to further filter query results.

                               Note: Currently, only MapServer, GPServer, and FeatureServer are the
                                     only supported service types.
        ------------------     --------------------------------------------------------------------
        source                 Optional String. The source of logged events.
        ------------------     --------------------------------------------------------------------
        stack_traces           Optional Boolean. If `True` the stack trace is returned.
        ------------------     --------------------------------------------------------------------
        out_folder             Optional string. The save folder location.
        ==================     ====================================================================

        """

        if out_folder is None:
            out_folder: str = tempfile.gettempdir()
        if (
            level
            and isinstance(level, str)
            and level.upper()
            in ["VERBOSE", "SEVERE", "WARNING", "INFO", "FINE", "DEBUG"]
        ):
            level = level.upper()
        else:
            level = ""
            _log.warning("`level` is not a valid value, ignoring")
        num: int = 10000
        if isinstance(users, list):
            users = ",".join(users)
        elif users and not isinstance(users, str):
            _log.warning("`users` is a string, accepting the value.")
        elif users and not isinstance(users, (str, list)):
            _log.warning("`users` is not a list[str] or string, ignoring the value.")
            users = ""
        if start_time and isinstance(start_time, _dt.datetime):
            start_time = start_time.strftime("%Y-%m-%dT%H:%M:%S")
        elif start_time:
            _log.warning("Invalid `start_time`, please use a datetime object")
        if end_time and isinstance(end_time, _dt.datetime):
            end_time = end_time.strftime("%Y-%m-%dT%H:%M:%S")
        elif end_time:
            _log.warning("Invalid `end_time`, please use a datetime object")
        url: str = f"{self._url}/export"
        if num > 10000:
            _log.warning(
                "The maximum value is 10,000 records.  Adjusting value to 10,000."
            )
            num = 10000
        params = {
            "logLevel": level or "",
            "startTime": start_time or "",
            "endTime": end_time or "",
            "logCode": log_code or "",
            "users": users or "",
            "requestIDs": request_ids or "",
            "serviceTypes": service_types or "",
            "source": source or "",
            "searchQuery": query or "",
            "sortBy": "bestMatch",
            "sortOrder": "desc",
            "num": num,
            "stackTraces": json.dumps(stack_traces),
            "f": "json",
        }
        resp = self._con.post(url, params)
        status_url: str = resp.get("jobsUrl", None)
        if status_url:
            data: dict[str, Any] = self._status_url(status_url)
            # download the result
            if "result" in data and "URL" in data["result"]:
                upload_url = data["result"]["URL"] + "/download"
                return self._con.get(
                    upload_url,
                    params={
                        "f": "json",
                    },
                    out_folder=out_folder,
                )
            else:
                return data

        else:
            return resp

    def _status_url(self, url) -> dict[str, Any]:
        import time

        resp = self._con.get(
            url,
            {
                "f": "json",
            },
        )
        i = 1
        # 'CANCELLING',
        while not resp.get("status", "FAILED") in [
            "COMPLETED",
            "FAILED",
            "CANCELLED",
            "TIME OUT",
        ]:
            time.sleep(i)
            resp = self._con.get(
                url,
                {
                    "f": "json",
                },
            )
            i += 1
            if i > 5:
                i = 5
        return resp

    # ----------------------------------------------------------------------
    def search(
        self,
        query: str,
        sort_by: str = "bestMatch",
        sort_order: str = "desc",
        show_stack: bool = False,
        return_count: bool = False,
    ) -> Union[int, List[Dict[str, Any]]]:
        """
        Allows after to search your organization's logs for specific log records.


        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        query                  Required String. The search terms used to query your organization's
                               logs. This parameter supports keywords (for example, completed) and
                               phrases (for example, completed successfully).
        ------------------     --------------------------------------------------------------------
        sort_by                Optional String. Specifies the way in which search results are
                               sorted. Sorting by `bestMatch` returns the results that best match
                               the `query` values. Sorting by `time` sorts the search results by
                               the time information recorded in the log's time stamp, the order
                               of which is determined by the `sort_order` parameter.

                               Values: `bestMatch` or `time`
        ------------------     --------------------------------------------------------------------
        sort_order             Optional String. The sort order for the results, either descending
                               or ascending. This parameter is ignored if sort_by is set to
                               `bestMatch`.

                               Values: `desc` or `asc`
        ------------------     --------------------------------------------------------------------
        show_stack             Optional Boolean. Specifies whether stack traces are included in the
                               search results. The default is false.
        ------------------     --------------------------------------------------------------------
        return_count           Optional Boolean.  If true, only returns a count of the logs that
                               would be returned by the search operation. If false, the response
                               includes the search results in full. The default is false.
        ==================     ====================================================================


        :return: If return_count==False, the messages as a list are returned, else an Integer

        """
        messages = []
        params = {
            "searchQuery": query,
            "sortBy": sort_by,
            "sortOrder": sort_order,
            "showStackTraces": show_stack,
            "countOnly": return_count,
            "num": 1000,
            "f": "json",
        }
        url = f"{self._url}/search"
        res = self._con.get(url, params)
        if return_count:
            return res.get("total", 0)
        messages.extend(res["messages"])
        while res["nextStart"] > -1:
            params["start"] = res["nextStart"]
            res = self._con.get(url, params)
            messages.extend(res["messages"])
        return messages

    # ----------------------------------------------------------------------
    def edit(self, level="WARNING"):
        """
        Provides log editing capabilities for the entire site.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        level                  Optional string. The log level.  Can be one of (in severity order):
                               OFF, DEBUG, VERBOSE, FINE, INFO, WARNING, SEVERE. The default is WARNING.
        ==================     ====================================================================


        :return: Boolean. True if successful else False.

        """
        url = self._url + "/settings/edit"
        allowed_levels = (
            "OFF",
            "SEVERE",
            "WARNING",
            "INFO",
            "FINE",
            "VERBOSE",
            "DEBUG",
        )
        current_settings = self.settings
        current_settings["f"] = "json"

        if level.upper() in allowed_levels:
            current_settings["logLevel"] = level.upper()

        res = self._con.post(path=url, postdata=current_settings)
        if "success" in res:
            return res["success"] == "true"
        return res

    # ----------------------------------------------------------------------
    def query(
        self,
        start_time=None,
        end_time=None,
        level="WARNING",
        log_code=None,
        users=None,
        request_ids=None,
        service_types=None,
        source=None,
        show_stack_traces=True,
        num=1000,
    ):
        """
        The query operation on the logs resource provides a way to
        aggregate, filter, and page through logs across the entire site.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        start_time             Optional string. The oldest time to query logs against, formatted as
                               either a timestamp (yyyy-mm-ddThh:mm:ss) or milliseconds from epoch.
                               The default is the beginning of all recorded logs.
        ------------------     --------------------------------------------------------------------
        end_time               Optional string. The most recent time to query against, formatted as
                               either a timestamp (yyyy-mm-ddThh:mm:ss) or milliseconds from epoch.
                               The default is the current date.
        ------------------     --------------------------------------------------------------------
        level                  Optional string. Gets only the records with a log level at or more
                               severe than the level declared here. Can be one of (in severity
                               order): DEBUG, VERBOSE, FINE, INFO, WARNING, SEVERE. The
                               default is WARNING.
        ------------------     --------------------------------------------------------------------
        log_code               Optional String. Specifies the log codes assigned to the logs. When
                               specified, query will return logs associated with those codes.
        ------------------     --------------------------------------------------------------------
        users                  Optional String. The username of a user within the organization that
                               can be used to further filter log results.
        ------------------     --------------------------------------------------------------------
        request_ids            Optional String. An ID assigned to a specific server event.
        ------------------     --------------------------------------------------------------------
        service_types          Optional String. The service type of a service within the
                               organization that can be used to further filter query results.

                               Note: Currently, only MapServer, GPServer, and FeatureServer are the
                                     only supported service types.
        ------------------     --------------------------------------------------------------------
        source                 Optional String. The source of logged events.
        ------------------     --------------------------------------------------------------------
        show_stack_traces      Optional Boolean. If `True` the stack trace is returned.
        ------------------     --------------------------------------------------------------------
        num                    Optional Int.  The maximum number of log records to be returned by
                               this query. The default messages per page is 1000. The limit for
                               this parameter is 10000 records.
        ==================     ====================================================================

        :return:
           A JSON of the log items that match the query. If export option is set to True, the
           output log file path is returned.

        """

        allowed_levels = (
            "SEVERE",
            "WARNING",
            "INFO",
            "FINE",
            "VERBOSE",
            "DEBUG",
        )
        params = {"f": "json", "num": num}
        params["start"] = 1
        url = "{url}/query".format(url=self._url)
        if start_time is not None and isinstance(start_time, datetime):
            params["startTime"] = start_time.strftime("%Y-%m-%dT%H:%M:%S,%f")
        if end_time is not None and isinstance(end_time, datetime):
            params["endTime"] = end_time.strftime("%Y-%m-%dT%H:%M:%S,%f")
        if level.upper() in allowed_levels:
            params["logLevel"] = level
        if log_code is not None:
            params["logCode"] = log_code
        if users:
            params["users"] = users
        if request_ids:
            params["requestIDs"] = request_ids
        if service_types:
            params["serviceTypes"] = service_types
        if source:
            params["source"] = source
        if show_stack_traces is not None:
            params["showStackTraces"] = show_stack_traces
        res = self._con.get(url, params)
        messages = res["messages"]
        # ['messages', 'total', 'start', 'num', 'nextStart', 'query']
        while res["nextStart"] != -1:
            params["start"] = res["nextStart"]
            res = self._con.get(url, params)
            messages.extend(res["messages"])
        return messages



import tensorflow as tf
from tensorflow.keras.layers import (
    Input,
    Conv2D,
    Dropout,
    ReLU,
    BatchNormalization,
    UpSampling2D,
    Reshape,
    Layer,
    AveragePooling2D,
    MaxPool2D,
    GlobalAveragePooling2D,
    GlobalMaxPool2D,
    Concatenate,
    Flatten,
    Dense,
    ZeroPadding2D,
    Cropping2D,
    SeparableConv2D,
)
from tensorflow.keras import Model
from .._utils.common_tf import (
    get_padding_info,
    get_cropping_info,
    get_skip_connection_info,
    get_channel_axis,
    UpSample2DToSize,
)


def get_unet_tf_head_output(
    backbone_initalized,
    data_format,
    n_output_classes,
    skip_lyr_idxs,
    skip_lyr_channels,
    mobile_optimized=False,
):
    # Get channel axis
    ca = channel_axis = get_channel_axis(data_format=data_format)

    # Mobile Optimized convolutions
    bottle_bottleneck_ratio = 2
    bottle_out_ratio = 1
    if mobile_optimized:
        conv_obj = SeparableConv2D
        bottle_bottleneck_ratio = 0.7
        bottle_out_ratio = 0.5
    else:
        conv_obj = Conv2D

    ### Unet Middle Block - Bottleneck ###
    # conv block 1
    bottle_channels = backbone_initalized.output.shape[ca]
    bottle_block_conv1_bn0 = BatchNormalization(
        axis=channel_axis, name=f"bottle_block_conv1_bn0"
    )(backbone_initalized.output)
    bottle_block_conv1 = conv_obj(
        round(bottle_channels * bottle_bottleneck_ratio),
        3,
        data_format=data_format,
        padding="same",
        activation="relu",
        name=f"bottle_block_conv1",
    )(bottle_block_conv1_bn0)
    bottle_block_conv1_bn1 = BatchNormalization(
        axis=channel_axis, name=f"bottle_block_conv1_bn1"
    )(bottle_block_conv1)

    # conv block 3
    bottle_block_conv3 = conv_obj(
        round(bottle_channels * bottle_out_ratio),
        3,
        data_format=data_format,
        padding="same",
        activation="relu",
        name=f"bottle_block_conv3",
    )(
        bottle_block_conv1_bn1
    )  # (bottle_block_conv2_drop1)
    bottle_block_conv3_bn1 = BatchNormalization(
        axis=channel_axis, name=f"bottle_block_conv3_bn1"
    )(bottle_block_conv3)
    bottle_block_conv3_drop1 = Dropout(0.3, name="bottle_block_conv3_drop1")(
        bottle_block_conv3_bn1
    )

    block_output_layer = bottle_block_conv3_drop1

    ### Decoder Block ###
    for i, idx in enumerate(skip_lyr_idxs):
        # decoder block i

        skip_con_output = backbone_initalized.layers[idx].output
        block_in_channels = block_output_layer.shape[ca]
        conv1_channels = round(block_in_channels / 2)
        conv2_channels = conv3_channels = round(
            (block_in_channels + skip_lyr_channels[i]) / 2
        )
        if idx == skip_lyr_idxs[-1]:
            conv_obj = Conv2D
            conv1_channels = block_in_channels
            if not mobile_optimized:
                conv2_channels = round(conv2_channels / 2)
        else:
            if mobile_optimized:
                conv2_channels = round((conv1_channels + skip_lyr_channels[i]) / 2)
        skip_con_size = skip_con_output.shape[2]

        # make input channels equal to target
        decoder_block_conv1 = SeparableConv2D(
            conv1_channels,
            3,
            data_format=data_format,
            padding="same",
            activation="relu",
            name=f"decoder_block{i}_conv1",
        )(block_output_layer)
        decoder_block_bn1 = BatchNormalization(
            axis=channel_axis, name=f"decoder_block{i}_bn1"
        )(decoder_block_conv1)

        # Resize input
        decoder_block_upsample1 = UpSample2DToSize(
            skip_con_size, name=f"decoder_block{i}_resize1"
        )(decoder_block_bn1)

        # Concatenate data from skip connection
        decoder_block_concat1 = Concatenate(
            axis=channel_axis, name=f"decoder_block{i}_concat1"
        )([decoder_block_upsample1, skip_con_output])

        # first conv on concatenated data
        decoder_block_conv2 = conv_obj(
            conv2_channels,
            3,
            data_format=data_format,
            padding="same",
            activation="relu",
            name=f"decoder_block{i}_conv2",
        )(decoder_block_concat1)
        decoder_block_bn2 = BatchNormalization(
            axis=channel_axis, name=f"decoder_block{i}_bn2"
        )(decoder_block_conv2)

        if idx == skip_lyr_idxs[-1]:
            block_output_layer = decoder_block_bn2
        else:
            # second conv on concatenated data
            decoder_block_conv3 = conv_obj(
                conv3_channels,
                3,
                data_format=data_format,
                padding="same",
                activation="relu",
                name=f"decoder_block{i}_conv3",
            )(decoder_block_bn2)
            decoder_block_bn3 = BatchNormalization(
                axis=channel_axis, name=f"decoder_block{i}_bn3"
            )(decoder_block_conv3)
            block_output_layer = decoder_block_bn3
        #
    decoder_block_output = block_output_layer

    # Final Classification Conv
    class_block_drop1 = Dropout(0.3, name="class_block_drop1")(decoder_block_output)
    class_block_conv1 = Conv2D(
        n_output_classes,
        3,
        data_format=data_format,
        padding="same",
        activation=None,
        name=f"class_block_conv1",
    )(class_block_drop1)

    # Cropping Block ##
    top_crop, bottom_crop, left_crop, right_crop = get_cropping_info(
        decoder_block_output, backbone_initalized.input, data_format=data_format
    )
    crop_block_crop1 = Cropping2D(
        cropping=((top_crop, bottom_crop), (left_crop, right_crop)),
        data_format=data_format,
        name=f"crop_block_crop1",
    )(class_block_conv1)

    model_output_layer = crop_block_crop1

    return model_output_layer


def get_unet_tf_model(backbone_initalized, data, mobile_optimized=False):
    """
    creates a U-net model from the supplied backbone
    that can be used with the supplied data.
    """
    # Get Data Format
    data_format = (
        tf.keras.backend.image_data_format()
    )  #'channels_first' or 'channels_last'

    # get skip connection information
    skip_lyr_idxs, skip_lyr_channels = get_skip_connection_info(
        backbone_initalized, data_format=data_format
    )

    # get head output
    head_output_layer = get_unet_tf_head_output(
        backbone_initalized=backbone_initalized,
        data_format=data_format,
        n_output_classes=data.c,
        skip_lyr_idxs=skip_lyr_idxs,
        skip_lyr_channels=skip_lyr_channels,
        mobile_optimized=mobile_optimized,
    )

    # Return Model
    return Model(inputs=backbone_initalized.input, outputs=head_output_layer)


def analyze_pred_unet_tf(activations):
    return tf.argmax(activations, axis=get_channel_axis())



from __future__ import annotations
from arcgis.gis import GIS
from typing import TypeVar

T = TypeVar("T")
V = TypeVar("V")


class RuntimeManager:
    """
    Provides information about the Runtimes in the Notebook Server

    ================  ===============================================================================
    **Parameter**      **Description**
    ----------------  -------------------------------------------------------------------------------
    url               Required String. The base url for the RuntimeManager endpoints.
    ----------------  -------------------------------------------------------------------------------
    gis               Required GIS. The ArcGIS Online connection object.
    ================  ===============================================================================

    """

    def __init__(self, url: str, gis: GIS):
        """initializer"""
        self._url = url
        self._gis = gis

    @property
    def properties(self) -> dict[T, V]:
        """
        Returns the runtimes on the GIS

        :returns: dict[T,V]
        """
        params = {"f": "json"}
        return self._gis._con.get(self._url, params)

    def list(self) -> list[dict[T, V]]:
        """
        returns a list of runtimes on the system

        :returns: list[dict[T,V]]
        """
        params = {"f": "json"}
        return self._gis._con.get(self._url, params).get("runtimes", [])

    def manifest(self, id: str) -> dict[T, V]:
        """
        returns a dictionary containing all the libraries for that runtime

        :returns: dict[T,V]
        """
        url = f"{self._url}/{id}/manifest"
        params = {"f": "json"}
        return self._gis._con.get(url, params)



import logging

log = logging.getLogger()

from arcgis.widgets._mapview._raster.engine import RasterData, MatplotlibRasterEngine


class RasterIOEngine(MatplotlibRasterEngine):
    def try_get_raster_data(self, image, bands) -> RasterData:
        try:
            return self._try_get_raster_data(image, bands)
        except Exception as e:
            return super().try_get_raster_data(image, bands)

    def _try_get_raster_data(self, image, bands) -> RasterData:
        import rasterio

        with rasterio.open(image) as dataset:
            extent = {
                "xmin": dataset.bounds.left,
                "ymin": dataset.bounds.bottom,
                "xmax": dataset.bounds.right,
                "ymax": dataset.bounds.top,
                "spatialReference": {},
            }
            if dataset.crs and dataset.crs.to_epsg():
                extent["spatialReference"]["wkid"] = dataset.crs.to_epsg()
            elif dataset.crs and dataset.crs.to_wkt():
                extent["spatialReference"]["wkt"] = dataset.crs.to_wkt()
            else:
                extent = {}
            img_data = dataset.read()
            img_data = img_data.transpose((1, 2, 0)).squeeze()
            if bands:
                img_data = img_data[:, :, bands].squeeze()
            return RasterData(img_data=img_data, extent=extent, id_=str(hash(image)))
